from .schema import *
from .writer import Writer
from .enum import gen_enum
from .struct import gen_struct
from .function import gen_function
from .converters import is_vmath_type, has_vmath_converter, set_enum_converter

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .meta import Header

class Library:
    def __init__(self, name: str) -> None:
        self.name = name
        # ['defines', 'structs', 'aliases', 'enums', 'callbacks', 'functions']
        self.structs = []   # type: list[Struct]
        self.aliases = {}   # type: dict[str, str | c_ast.Node]
        self.enums = []     # type: list[Enum]
        self.functions = [] # type: list[Function]
        self.callbacks = set() # type: set[str]

    def unalias(self, name: str) -> c_ast.Node | str:
        while name in self.aliases:
            node = self.aliases[name]
            if isinstance(node, str):
                name = node
            else:
                return node
        return name

    def build(self, *, glue_dir='.', stub_dir='.', includes: list[str] | None = None):
        self.remove_unsupported()

        for k in self.aliases.keys():
            node = self.unalias(k)
            if isinstance(node, c_ast.Enum):
                set_enum_converter(k)

        w, pyi_w = Writer(), Writer()

        if has_vmath_converter():
            pyi_w.write('from vmath import vec2, vec3, vec2i, vec3i, mat3x3, color32')
        pyi_w.write('from typing import overload')
        pyi_w.write('from stdc import intptr')
        pyi_w.write('')

        w.write('#include "pocketpy.h"')
        w.write(f'#include <string.h>')
        w.write(f'#include <stdint.h>')
        
        if includes:
            for include in includes:
                w.write(f'#include "{include}"')

        w.write('')
        w.write('#define ADD_ENUM(name) py_newint(py_emplacedict(mod, py_name(#name)), name)')
        w.write('')

        for k in self.aliases.keys():
            node = self.unalias(k)
            if isinstance(node, str):
                w.write(f'#define tp_user_{k} tp_user_{node}')
        w.write('')

        reg_exprs = [
            gen_struct(w, pyi_w, struct)
            for struct in self.structs
        ]

        w.write('/* functions */')
        for function in self.functions:
            gen_function(w, pyi_w, function)
        
        w.write(f'void py__add_module_{self.name}() {{')
        w.indent()

        w.write(f'py_GlobalRef mod = py_newmodule("{self.name}");')

        w.write('/* structs */')
        for reg_expr in reg_exprs:
            w.write(reg_expr)

        w.write('/* aliases */')
        pyi_w.write('# aliases')
        for k in self.aliases.keys():
            node = self.unalias(k)
            if isinstance(node, str):
                w.write(f'py_setdict(mod, py_name("{k}"), py_getdict(mod, py_name("{node}")));')
                pyi_w.write(f'{k} = {node}')
            elif isinstance(node, c_ast.Enum):
                w.write(f'py_setdict(mod, py_name("{k}"), py_tpobject(tp_int));')
                pyi_w.write(f'{k} = int')

        w.write('/* functions */')
        for function in self.functions:
            w.write(f'py_bindfunc(mod, "{function.name}", &cfunc__{function.name});')

        w.write('/* enums */')
        pyi_w.write('# enums')
        for enum in self.enums:
            gen_enum(w, pyi_w, enum)
        w.dedent()
        w.write('}')
        with open(f'{glue_dir}/{self.name}.c', 'w') as f:
            f.write(str(w))
        with open(f'{stub_dir}/{self.name}.pyi', 'w') as f:
            f.write(str(pyi_w))

    def remove_unsupported(self):
        functions = []
        for f in self.functions:
            if f.params and f.params[-1].type == '...':
                print('[WARN]', f.signature(), 'is variadic')
                continue
            for p in f.params:
                if p.type in self.callbacks:
                    print('[WARN]', f.signature(), 'has callback param')
                    break
            else:
                functions.append(f)
        self.functions.clear()
        self.functions.extend(functions)
                
    @staticmethod
    def from_raylib(data: dict):
        self = Library('raylib')
        for struct in data['structs']:
            name = struct['name']
            if is_vmath_type(name):
                print(f'[INFO] {name} is a vmath type, skipping')
                continue
            self.structs.append(Struct(
                name=struct['name'],
                typedef_name=struct['name'],
                desc=struct['description'],
                fields=[StructField(
                    type=field['type'],
                    name=field['name'],
                    desc=field['description']
                ) for field in struct['fields']]
            ))
        for alias in data['aliases']:
            self.aliases[alias['name']] = str(alias['type'])
        for enum in data['enums']:
            self.enums.append(Enum(
                name=enum['name'],
                desc=enum['description'],
                values=[EnumValue(
                    name=value['name'],
                    value=value['value'],
                    desc=value['description']
                ) for value in enum['values']]
            ))
        for function in data['functions']:
            self.functions.append(Function(
                name=function['name'],
                desc=function['description'],
                params=[FunctionParam(
                    type=param['type'],
                    name=param['name']
                ) for param in function['params']
                ] if 'params' in function else [],
                ret_type=function['returnType']
            ))
        for callback in data['callbacks']:
            self.callbacks.add(callback['name'])
        return self
    
    @staticmethod
    def from_header(name: str, header: 'Header'):
        from ffigen.meta import schema
        self = Library(name)
        for type in header.types:
            if isinstance(type, schema.NamedFields):
                if type.is_opaque():
                    continue
                else:
                    fields = type.fields
                    assert fields is not None
                    self.structs.append(Struct(
                        name=type.name,
                        typedef_name=type.typedef_name,
                        fields=[StructField(
                            type=field_type,
                            name=field_name
                        ) for field_name, field_type in fields.items()]
                    ))
            elif isinstance(type, schema.Enum):
                self.enums.append(Enum(
                    name=type.name,
                    values=[EnumValue(
                        name=value,
                        value=None
                    ) for value in type.values]
                ))
        for k, v in header.type_aliases.items():
            self.aliases[k] = v

        for function in header.functions:
            self.functions.append(Function(
                name=function.name,
                params=[FunctionParam(
                    type=param[0],
                    name=param[1] or f'_{i}'
                ) for i, param in enumerate(function.args)],
                ret_type=function.ret
            ))
        return self


import re
import shutil
import os
import sys
import time
from typing import List, Dict

assert os.system("python prebuild.py") == 0

ROOT = 'include/pocketpy'
PUBLIC_HEADERS = ['config.h', 'export.h', 'vmath.h', 'pocketpy.h']

COPYRIGHT = '''/*
 *  Copyright (c) 2026 blueloveTH
 *  Distributed Under The MIT License
 *  https://github.com/pocketpy/pocketpy
 */
 '''

def read_file(path):
	with open(path, 'rt', encoding='utf-8') as f:
		return f.read()

def write_file(path, content):
	with open(path, 'wt', encoding='utf-8', newline='\n') as f:
		f.write(content)
	
if os.path.exists('amalgamated'):
	shutil.rmtree('amalgamated')
	time.sleep(0.5)

os.mkdir('amalgamated')

class Header:
	path: str
	content: str		# header source
	dependencies: List[str]

	def __init__(self, path: str):
		self.path = path
		self.dependencies = []
		self.content = read_file(f'{ROOT}/{path}')

		# process raw content and get dependencies
		self.content = self.content.replace('#pragma once', '')
		def _replace(m):
			path = m.group(1)
			if path.startswith('xmacros/'):
				return read_file(f'{ROOT}/{path}') + '\n'
			if path in PUBLIC_HEADERS:
				return ''	# remove include
			if path != self.path:
				self.dependencies.append(path)
			return ''	# remove include
		
		self.content = re.sub(
			r'#include\s+"pocketpy/(.+)"\s*',
			_replace,
			self.content
		)

	def __repr__(self):
		return f'Header({self.path!r}, dependencies={self.dependencies})'
	
	def text(self):
		return f'// {self.path}\n{self.content}\n'


headers: Dict[str, Header] = {}

for entry in os.listdir(ROOT):
	if os.path.isdir(f'{ROOT}/{entry}'):
		if entry == 'xmacros' or entry in PUBLIC_HEADERS:
			continue
		files = os.listdir(f'{ROOT}/{entry}')
		for file in sorted(files):
			assert file.endswith('.h')
			if entry in PUBLIC_HEADERS:
				continue
			headers[f'{entry}/{file}'] = Header(f'{entry}/{file}')

def merge_c_files():
	c_files = [
		COPYRIGHT,
		'\n',
		'#define PK_IS_AMALGAMATED_C',
		'\n',
		'#include "pocketpy.h"',
		'\n'
	]

	# merge internal headers
	internal_h = []
	while True:
		for h in headers.values():
			if not h.dependencies:
				break
		else:
			if headers:
				print(headers)
				raise RuntimeError("Circular dependencies detected")
			break
		# print(h.path)
		internal_h.append(h.text())
		del headers[h.path]
		for h2 in headers.values():
			h2.dependencies = [d for d in h2.dependencies if d != h.path]

	c_files.extend(internal_h)

	def _replace(m):
		path = m.group(1)
		if path.startswith('xmacros/'):
			return read_file(f'{ROOT}/{path}') + '\n'
		return ''	# remove include

	for root, _, files in os.walk('src/'):
		for file in files:
			if file.endswith('.c'):
				path = os.path.join(root, file)
				c_files.append(f'// {path}\n')
				content = read_file(path)
				content = re.sub(
					r'#include\s+"pocketpy/(.+)"\s*',
					_replace,
					content,
				)
				c_files.append(content)
				c_files.append('\n')
	return ''.join(c_files)

def merge_h_files():
	h_files = [
		COPYRIGHT,
		'#pragma once',
		'\n',
		'#define PK_IS_PUBLIC_INCLUDE',
	]

	def _replace(m):
		path = m.group(1)
		if path.startswith('xmacros/'):
			return read_file(f'{ROOT}/{path}') + '\n'
		return ''	# remove include
	
	for path in PUBLIC_HEADERS:
		content = read_file(f'{ROOT}/{path}')
		content = content.replace('#pragma once', '')
		content = re.sub(
					r'#include\s+"pocketpy/(.+)"\s*',
					_replace,
					content,
				)
		h_files.append(content)
	return '\n'.join(h_files)


write_file('amalgamated/pocketpy.c', merge_c_files())
write_file('amalgamated/pocketpy.h', merge_h_files())

shutil.copy("src2/main.c", "amalgamated/main.c")

def checked_sh(cmd):
	ok = os.system(cmd)
	assert ok == 0, f"command failed: {cmd}"

if sys.platform in ['linux', 'darwin']:
	common_flags = "-O1 --std=c11 -lm -ldl -lpthread -Iamalgamated"
	checked_sh(f"gcc -o main amalgamated/pocketpy.c src2/example.c {common_flags}")
	checked_sh("./main && rm -f ./main")
	checked_sh(f"gcc -o main amalgamated/pocketpy.c amalgamated/main.c {common_flags}")


print("amalgamated/pocketpy.h")



from vmath import vec3

def linear_srgb_to_srgb(rgb: vec3) -> vec3: ...
def srgb_to_linear_srgb(rgb: vec3) -> vec3: ...
def srgb_to_hsv(rgb: vec3) -> vec3: ...
def hsv_to_srgb(hsv: vec3) -> vec3: ...
def oklch_to_linear_srgb(lch: vec3) -> vec3: ...
def linear_srgb_to_oklch(rgb: vec3) -> vec3: ...



from dataclasses import dataclass
from pycparser import c_ast

@dataclass
class StructField:
    type: str
    name: str
    desc: str | None = None

@dataclass
class EnumValue:
    name: str
    value: int | None
    desc: str | None = None

@dataclass
class Struct:
    name: str | None = None
    typedef_name: str | None = None
    desc: str | None = None
    fields: list[StructField] | None = None

    @property
    def code_name(self):
        if self.typedef_name:
            return self.typedef_name
        assert self.name is not None
        return f'struct {self.name}'
    
    @property
    def identifier(self):
        if self.name:
            return self.name
        assert self.typedef_name is not None
        return self.typedef_name

@dataclass
class Enum:
    name: str
    values: list[EnumValue]
    desc: str | None = None

@dataclass
class FunctionParam:
    type: str
    name: str

@dataclass
class Function:
    name: str
    params: list[FunctionParam]
    ret_type: str
    desc: str | None = None

    def signature(self) -> str:
        return f'{self.ret_type} {self.name}({", ".join([f"{param.type} {param.name}" for param in self.params])})'




def all(iterable):
    for i in iterable:
        if not i:
            return False
    return True

def any(iterable):
    for i in iterable:
        if i:
            return True
    return False

def enumerate(iterable, start=0):
    n = start
    for elem in iterable:
        yield n, elem
        n += 1

def __minmax_reduce(op, args):
    if len(args) == 2:  # min(1, 2)
        return args[0] if op(args[0], args[1]) else args[1]
    if len(args) == 0:  # min()
        raise TypeError('expected 1 arguments, got 0')
    if len(args) == 1:  # min([1, 2, 3, 4]) -> min(1, 2, 3, 4)
        args = args[0]
    args = iter(args)
    try:
        res = next(args)
    except StopIteration:
        raise ValueError('args is an empty sequence')
    while True:
        try:
            i = next(args)
        except StopIteration:
            break
        if op(i, res):
            res = i
    return res

def min(*args, key=None):
    key = key or (lambda x: x)
    return __minmax_reduce(lambda x,y: key(x)<key(y), args)

def max(*args, key=None):
    key = key or (lambda x: x)
    return __minmax_reduce(lambda x,y: key(x)>key(y), args)

def sum(iterable):
    res = 0
    for i in iterable:
        res += i
    return res

def map(f, iterable):
    for i in iterable:
        yield f(i)

def filter(f, iterable):
    for i in iterable:
        if f(i):
            yield i

class zip:
    def __init__(self, *iterables):
        self.iterables = [iter(it) for it in iterables]

    def __iter__(self):
        return self

    def __next__(self):
        result = []
        for it in self.iterables:
            result.append(next(it))
        return tuple(result)

def reversed(iterable):
    a = list(iterable)
    a.reverse()
    return a

def sorted(iterable, key=None, reverse=False):
    a = list(iterable)
    a.sort(key=key, reverse=reverse)
    return a


def help(obj):
    if hasattr(obj, '__func__'):
        obj = obj.__func__
    # print(obj.__signature__)
    if obj.__doc__:
        print(obj.__doc__)

def complex(real, imag=0):
    import cmath
    return cmath.complex(real, imag) # type: ignore

def dir(obj) -> list[str]:
    tp_module = type(__import__('math'))
    if isinstance(obj, tp_module):
        return [k for k, _ in obj.__dict__.items()]
    names = set()
    if not isinstance(obj, type):
        obj_d = obj.__dict__
        if obj_d is not None:
            names.update([k for k, _ in obj_d.items()])
        cls = type(obj)
    else:
        cls = obj
    while cls is not None:
        names.update([k for k, _ in cls.__dict__.items()])
        cls = cls.__base__
    return sorted(list(names))

class set:
    def __init__(self, iterable=None):
        iterable = iterable or []
        self._a = {}
        self.update(iterable)

    def add(self, elem):
        self._a[elem] = None
        
    def discard(self, elem):
        self._a.pop(elem, None)

    def remove(self, elem):
        del self._a[elem]
        
    def clear(self):
        self._a.clear()

    def update(self, other):
        for elem in other:
            self.add(elem)

    def __len__(self):
        return len(self._a)
    
    def copy(self):
        return set(self._a.keys())
    
    def __and__(self, other):
        return {elem for elem in self if elem in other}

    def __sub__(self, other):
        return {elem for elem in self if elem not in other}
    
    def __or__(self, other):
        ret = self.copy()
        ret.update(other)
        return ret

    def __xor__(self, other): 
        _0 = self - other
        _1 = other - self
        return _0 | _1

    def union(self, other):
        return self | other

    def intersection(self, other):
        return self & other

    def difference(self, other):
        return self - other

    def symmetric_difference(self, other):      
        return self ^ other
    
    def __eq__(self, other):
        if not isinstance(other, set):
            return NotImplemented
        return len(self ^ other) == 0
    
    def __ne__(self, other):
        if not isinstance(other, set):
            return NotImplemented
        return len(self ^ other) != 0

    def isdisjoint(self, other):
        return len(self & other) == 0
    
    def issubset(self, other):
        return len(self - other) == 0
    
    def issuperset(self, other):
        return len(other - self) == 0

    def __contains__(self, elem):
        return elem in self._a
    
    def __repr__(self):
        if len(self) == 0:
            return 'set()'
        return '{'+ ', '.join([repr(i) for i in self._a.keys()]) + '}'
    
    def __iter__(self):
        return iter(self._a.keys())